/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <mpc8xx.h>
#include "flash_prv.h"

#define FLASH_ROM	0xFFFD
#define FLASH_RAM	0xFFFE
#define FLASH_MAN_UNKNOWN 0xFFFF0000

/* #define FLASH_ID_OVERRIDE */

flash_info_t	flash_info[CFG_MAX_FLASH_BANKS]; /* info for FLASH chips	*/

#if defined(CFG_ENV_IS_IN_FLASH)
# ifndef  CFG_ENV_ADDR
#  define CFG_ENV_ADDR	(CFG_FLASH_BASE + CFG_ENV_OFFSET)
# endif
# ifndef  CFG_ENV_SIZE
#  define CFG_ENV_SIZE	CFG_ENV_SECT_SIZE
# endif
# ifndef  CFG_ENV_SECT_SIZE
#  define CFG_ENV_SECT_SIZE  CFG_ENV_SIZE
# endif
#endif

/* NOTE:
 * portwidth = total width of flash bus in bytes, e.g. 4x8, 2x16 == portwidth 4
 * chipwidth = width of single chip in bytes, e.g. 1x6, 2x16 == chipwidth 2
 */

/*-----------------------------------------------------------------------
 * Functions
 */
static ulong flash_get_size (vu_char *addr, flash_info_t *info);
static int write_word (flash_info_t *info, ulong dest, ulong data);
static void flash_get_offsets (ulong base, flash_info_t *info);

/*-----------------------------------------------------------------------
 */

static unsigned int map_bank(int cs, int portwidth, ulong base, ulong size)
{
    	int i;
	ulong or_am=0, br_ps=0;
	volatile immap_t     *immap  = (immap_t *)CFG_IMMR;
	volatile memctl8xx_t *memctl = &immap->im_memctl;
	volatile unsigned int *brm[]={
		&memctl->memc_br0, &memctl->memc_br1,
		&memctl->memc_br2, &memctl->memc_br3,
		&memctl->memc_br4, &memctl->memc_br5,
		&memctl->memc_br6, &memctl->memc_br7};
	volatile unsigned int *orm[]={
		&memctl->memc_or0, &memctl->memc_or1,
		&memctl->memc_or2, &memctl->memc_or3,
		&memctl->memc_or4, &memctl->memc_or5,
		&memctl->memc_or6, &memctl->memc_or7};
	volatile unsigned int *br=brm[cs], *or=orm[cs];

	if(cs<=0 || cs>7) return 0;	/* don't allow remapping of CS 0 */

	if(base & 0x7ffff) return -1;	/* illegal base */

	if(size==0) {
	    /* unmap */
	    *br=0;*or=0;
	    return 0;
	}

	/* Compute the Address Mask */
	for (i=0; (size >> i) != 0; ++i) {};
	i--;

	if (size != (1 << i)) {
		printf ("## WARNING: FLASH size (0x%08lx = %ld MB) is not a power of 2\n",
			size, size>>20);
	}

	or_am=((unsigned long)~0) << i;

	switch(portwidth) {
		case 1: br_ps=BR_PS_8; break;
		case 2: br_ps=BR_PS_16; break;
		case 4: br_ps=BR_PS_32; break;
		default: return -1;
	}

	(*br) &= ~(BR_BA_MSK | BR_PS_MSK | BR_V);	/* disable bank */
	/* Remap FLASH according to real size */
	(*or) |= OR_AM_MSK;	/* mask out all */
	(*or) &= or_am | ~OR_AM_MSK;
	/* write out base address, port size, and enable */
	(*br) |= (base | br_ps | BR_V);
	/* printf("br%d:or%d 0x%x:0x%x\n", cs, cs, *br, *or); */
	return 0;
}

unsigned long flash_init (void)
{
	unsigned long total_size;
	unsigned long ext_flash_base=CFG_EXT_FLASH_BASE;
	unsigned long ext_flash_last_size=0x8000000;
	int i;

	/* Init: no FLASHes known */
	for (i=0; i < CFG_MAX_FLASH_BANKS; ++i)
	{
		int pwmap[]=CFG_FLASH_PORTWIDTH_MAP;
		int cwmap[]=CFG_FLASH_CHIPWIDTH_MAP;
		flash_info[i].flash_id = FLASH_UNKNOWN;
		flash_info[i].portwidth = pwmap[i];
		flash_info[i].chipwidth = cwmap[i];
		flash_info[i].size = 0;
	}

	total_size = 0;

	for (i=0; i < CFG_MAX_FLASH_BANKS; ++i)
	{
		unsigned long size=0;
		const int map[]=CFG_FLASH_CS_MAP;
		ulong base=CFG_FLASH_BASE;

		/* if cs0, use CFG_FLASH_BASE, fill rest starting from
		 * CFG_EXT_FLASH_BASE */
		if(map[i]!=0) {
		    base=ext_flash_base;
		}

		/* map it so we can see it */
		map_bank(map[i], flash_info[i].portwidth, base, ext_flash_last_size);
		size = flash_get_size((vu_char *)base, &flash_info[i]);

		if (flash_info[i].flash_id == FLASH_UNKNOWN)
		{
			printf ("## Unknown FLASH on Bank %d - Size = 0x%08lx = %ld MB\n", i, size, size>>20);
		}

		/* shrink to size */
		map_bank(map[i], flash_info[i].portwidth, base, size);

		if(map[i]!=0) {
		    ext_flash_base += size;
		    ext_flash_last_size = size;
		}

		total_size += size;
	}

	for (i=0; i < CFG_MAX_FLASH_BANKS && flash_info[i].size != 0; ++i)
	{
#if CFG_MONITOR_BASE >= CFG_FLASH_BASE
		/* monitor protection ON by default */
		flash_protect(FLAG_PROTECT_SET,
			      CFG_MONITOR_BASE,
			      CFG_MONITOR_BASE+monitor_flash_len-1,
			      &flash_info[i]);
#endif

#ifdef	CFG_ENV_IS_IN_FLASH
		/* ENV protection ON by default */
		flash_protect(FLAG_PROTECT_SET,
			      CFG_ENV_ADDR,
			      CFG_ENV_ADDR+CFG_ENV_SIZE-1,
			      &flash_info[i]);
#endif
	}

	return (total_size);
}

/*-----------------------------------------------------------------------
 */
static void flash_get_offsets (ulong base, flash_info_t *info)
{
	int i;
	/* set up sector start address table */
#if 0
	if ((info->flash_id & FLASH_TYPEMASK) == FLASH_AM040 ||
	    (info->flash_id & FLASH_TYPEMASK) == FLASH_AM080 ||
	    (info->flash_id & FLASH_TYPEMASK) == FLASH_ROM ||
	    (info->flash_id & FLASH_TYPEMASK) == FLASH_RAM ) {
		/* set sector offsets for uniform sector type	*/
		for (i = 0; i < info->sector_count; i++) {
			info->start[i] = base + (i * 0x00010000 * numchips);
		}
	} else {
		printf("%s: can't fill offsets\n", __FUNCTION__);
	}
#else
	int sectorsize=info->size/info->sector_count;

	if(info->sector_count<=CFG_MAX_FLASH_SECT) {
	    for (i = 0; i < info->sector_count; i++) {
		    info->start[i] = base + (i * sectorsize);
	    }
	} else {
	    printf("%s: can't fill offsets, too many sectors (%d>%d)\n",
		    __FUNCTION__, info->sector_count,CFG_MAX_FLASH_SECT);
	}
#endif
}

/*-----------------------------------------------------------------------
 */
void flash_print_info  (flash_info_t *info)
{
	int i;

	if (info->flash_id == FLASH_UNKNOWN)
	{
		printf ("missing or unknown FLASH type\n");
		return;
	}

	switch (info->flash_id & FLASH_VENDMASK)
	{
		case FLASH_MAN_AMD:	printf ("AMD ");		break;
		case FLASH_MAN_FUJ:	printf ("FUJITSU ");		break;
		case FLASH_MAN_BM:	printf ("BRIGHT MICRO ");	break;
		case FLASH_MAN_INTEL:	printf ("INTEL ");	break;
		default:		printf ("Unknown Vendor ");	break;
	}

	switch (info->flash_id & FLASH_TYPEMASK)
	{
		case FLASH_AM040:	printf ("29F040 or 29LV040 (4 Mbit, uniform sectors)\n");
			break;
		case FLASH_AM080:	printf ("29F080 or 29LV080 (8 Mbit, uniform sectors)\n");
		                	break;
		case FLASH_AM400B:	printf ("AM29LV400B (4 Mbit, bottom boot sect)\n");
					break;
		case FLASH_AM400T:	printf ("AM29LV400T (4 Mbit, top boot sector)\n");
					break;
		case FLASH_AM800B:	printf ("AM29LV800B (8 Mbit, bottom boot sect)\n");
					break;
		case FLASH_AM800T:	printf ("AM29LV800T (8 Mbit, top boot sector)\n");
					break;
		case FLASH_AM160B:	printf ("AM29LV160B (16 Mbit, bottom boot sect)\n");
					break;
		case FLASH_AM160T:	printf ("AM29LV160T (16 Mbit, top boot sector)\n");
					break;
		case FLASH_AM320B:	printf ("AM29LV320B (32 Mbit, bottom boot sect)\n");
					break;
		case FLASH_AM320T:	printf ("AM29LV320T (32 Mbit, top boot sector)\n");
					break;
		case FLASH_28F320J3A:	printf ("INTEL28F320J3A (32 Mbit)\n");
					break;
		case FLASH_28F640J3A:	printf ("INTEL28F640J3A (64 Mbit)\n");
					break;
		case FLASH_28F128J3A:	printf ("INTEL28F128J3A (128 Mbit)\n");
					break;

		case FLASH_ROM:		printf ("ROM\n");
					break;
		case FLASH_RAM:		printf ("RAM\n");
					break;
		default:		printf ("Unknown Chip Type\n");
					break;
	}

	printf ("  Size: %ld MB in %d Sectors\n",info->size >> 20, info->sector_count);

	printf ("  Sector Start Addresses:");

	for (i=0; i<info->sector_count; ++i)
	{
		if ((i % 5) == 0)
		{
			printf ("\n   ");
		}

		printf (" %08lX%s",
			info->start[i],
			info->protect[i] ? " (RO)" : "     ");
	}

	printf ("\n");
	return;
}

/*-----------------------------------------------------------------------
 */

/*
 * The following code cannot be run from FLASH!
 */

static inline void flash_cmd(flash_info_t *info, volatile unsigned char *addr,
	int offset, unsigned char cmd)
{
        unsigned long cmd32=cmd;
    	/* 32 bit chipwidth unsupported */
	if(info->chipwidth==2) {	/* 16 bit mode */
		cmd32<<=8; /* move into msb */
		if(info->portwidth==4) { /* 2x16 */
			/* assuming chips are in 16 bit mode */
			cmd32=(cmd32<<16)|cmd32;
			*(volatile unsigned long *)(addr+offset*2)=cmd32;
		} else if(info->portwidth==2) {	/* 1x16 */
		    	/* printf("mw.w 0x%p 0x%04lx\n", addr+offset,cmd32); */
			*(volatile unsigned short *)(addr+offset)=cmd32;
		}
	} else if(info->chipwidth==1) { /* 8 bit mode */
		if(info->portwidth==4) { /* 4x8 */
			/* untested */
			cmd32=(cmd<<24)|(cmd<<16)|(cmd<<8)|cmd;
			*(volatile unsigned long *)(addr+offset*4)=cmd32;
		} else if(info->portwidth==2) { /* 2x8 */
			/* untested */
			unsigned short cmd16=(cmd<<8)|cmd;
			*(volatile unsigned short *)(addr+offset*2)=cmd16;
		} else if(info->portwidth==1) { /* 1x8 */
			*(volatile unsigned char *)(addr+offset)=cmd;
		}
	}
}

static inline int flash_cmp(flash_info_t *info, volatile unsigned char *addr,
	int offset, unsigned char data)
{
    	unsigned long data32=data;
    	/* 32 bit chipwidth unsupported */
	if(info->chipwidth==2) {	/* 16 bit mode */
	    	data32<<=8; /* move into msb */
		if(info->portwidth==4) { /* 2x16 */
			/* untested */
			data32=(data32<16)|data32;
			return (*(volatile unsigned long *)(addr+offset*2)==data32);
		} else if(info->portwidth==2) { /* 1x16 */
			/* untested */
			return (*(volatile unsigned char *)(addr+offset)==data32);
		}
	} else if(info->chipwidth==1) {	/* 8 bit mode */
		if(info->portwidth==4) { /* 4x8 */
			/* untested */
			unsigned long data32=(data<<24)|(data<<16)|(data<<8)|data;
			return (*(volatile unsigned long *)(addr+offset*4)==data32);
		} else if(info->portwidth==2) { /* 2x8 */
			/* untested */
			unsigned short data16=(data<<8)|data;
			return (*(volatile unsigned short *)(addr+offset*2)==data16);
		} else if(info->portwidth==1) {	/* 1x8 */
			return (*(volatile unsigned char *)(addr+offset)==data);
		}
	}
	return 0;
}


static ulong flash_check_rom_ram(vu_char *addr, flash_info_t *info, uchar old[2])
{
	uchar save;
	ulong base = (ulong)addr;

	save = *addr;

	flash_cmd(info,addr,0,0xf0);
	flash_cmd(info,addr,0,0xf0);

	udelay(10);

	old[0] = addr[0];
	old[1] = addr[1];
	/* flash: data
	 * rom: data
	 * ram: 0xf0 */

	if(old[0]!=0xf0) {
	    /* flash or rom */
	    flash_cmd(info,addr,0,0xf0);
	    flash_cmd(info,addr,0,0xf0);

	    udelay(10);

	    if(*addr==0xf0) {
		/* this area is ROM */
		*addr=save;
#ifndef FLASH_ID_OVERRIDE
		info->flash_id = FLASH_ROM + FLASH_MAN_UNKNOWN;
		info->sector_count = 8;
		info->size = 0x80000;
#else
		info->flash_id = FLASH_MAN_AMD + FLASH_AM040;
		info->sector_count = 8;
		info->size = 0x80000;
#endif
		flash_get_offsets(base, info);
		return info->size;
	    }
	} else {
	    /* ram or rom/flash had 0xf0 in it */
	    *addr=0;

	    udelay(10);

	    if(*addr==0) {
		/* this area is RAM */
		*addr=save;
		info->flash_id = FLASH_RAM + FLASH_MAN_UNKNOWN;
		info->sector_count = 8;
		info->size = 0x80000;
		flash_get_offsets(base, info);
		return info->size;
	    }
	    flash_cmd(info,addr,0,0xf0);

	    udelay(10);
	}
	/* undetermined. check again after id and compare
	   with old[] */
	return 0;
}

static void get_manu_id(vu_char *addr, flash_info_t *info, int *manu, int *id)
{
        vu_char *caddr = (vu_char *)addr;
        vu_short *saddr = (vu_short *)addr;
        vu_long *laddr = (vu_long *)addr;

        switch(info->portwidth) {
            case 1:
                *manu = caddr[0];
                *manu |= (*manu)<<16;
                *id = caddr[1];
                break;
            case 2:
                *manu = saddr[0];
                *manu |= (*manu)<<16;
                *id = saddr[1];
                *id |= (*id)<<16;
                break;
            case 4:
                *manu = laddr[0];
                *id = laddr[1];
                break;
	}
}

static ulong flash_get_size (vu_char *addr, flash_info_t *info)
{
	short i;
	int id=-1,manu=-1;
	uchar old[2];
	int numchips;

	if(flash_check_rom_ram(addr,info,old)!=0) return info->size;

	if(info->chipwidth==0) info->chipwidth=1;	/* 8 bit mode */
	if(info->portwidth==0) info->portwidth=1;	/* 1x8 */

	if(info->portwidth<info->chipwidth) return info->size;

	numchips=info->portwidth/info->chipwidth;

	/* Write auto select command: read Manufacturer ID */
	flash_cmd(info, addr, 0x555, 0xaa);
	flash_cmd(info, addr, 0x2aa, 0x55);
	flash_cmd(info, addr, 0x555, 0x90);

	/* printf("md.b 0x%p 4\n%02x %02x %02x %02x\n", 
		addr, addr[0], addr[1], addr[2], addr[3]);
	 */

	if(old[0]==addr[0] && old[1]==addr[1]) {
#ifndef FLASH_ID_OVERRIDE
	    info->flash_id = FLASH_ROM + FLASH_MAN_UNKNOWN;
	    info->sector_count = 8;
	    info->size = 0x80000;
#else
	    info->flash_id = FLASH_MAN_AMD + FLASH_AM040;
	    info->sector_count = 8;
	    info->size = 0x80000;
	    info->chipwidth=1;
#endif
            flash_get_offsets((ulong)addr, info);
	    printf("ROM\n");
            return info->size;
	}
	
	get_manu_id(addr, info, &manu, &id);

	switch (manu)
	{
		case AMD_MANUFACT:
			info->flash_id = FLASH_MAN_AMD;
			break;

		case FUJ_MANUFACT:
			info->flash_id = FLASH_MAN_FUJ;
			break;

		case INTEL_MANUFACT:
			info->flash_id = FLASH_MAN_INTEL;
			break;

		default:
			printf("%s: unknown manu %x\n", __FUNCTION__,manu);
			info->flash_id = FLASH_UNKNOWN;
			info->sector_count = 0;
			info->size = 0;
			break;
	}

	switch (id)
	{
		case AMD_ID_F040B:
		case AMD_ID_LV040B:
			info->flash_id += FLASH_AM040;
			info->sector_count = 8;
			info->size = 0x00080000 * numchips;
			break;			/* 512 * 4 = 2MB	*/

		case AMD_ID_F080B:
			info->flash_id += FLASH_AM080;
			info->sector_count =16;
			info->size = 0x00100000 * numchips;
			break;			/* 1M * 4 = 4MB		*/

		case AMD_ID_LV400T:
			info->flash_id += FLASH_AM400T;
			info->sector_count = 11;
			info->size = 0x00040000 * numchips;
			break;			/* 256k * 4 = 1MB	*/

		case AMD_ID_LV400B:
			info->flash_id += FLASH_AM400B;
			info->sector_count = 11;
			info->size = 0x00040000 * numchips;
			break;			/* 256k * 4 = 1MB	*/

		case AMD_ID_LV800T:
			info->flash_id += FLASH_AM800T;
			info->sector_count = 19;
			info->size = 0x00080000 * numchips;
			break;			/* 512k * 4 = 2MB	*/

		case AMD_ID_LV800B:
			info->flash_id += FLASH_AM800B;
			info->sector_count = 19;
			info->size = 0x00080000 * numchips;
			break;			/* 512k * 4 = 2MB	*/

		case AMD_ID_LV160T:
			info->flash_id += FLASH_AM160T;
			info->sector_count = 35;
			info->size = 0x00100000 * numchips;
			break;			/* 1M * 4 = 4MB		*/

		case AMD_ID_LV160B:
			info->flash_id += FLASH_AM160B;
			info->sector_count = 35;
			info->size = 0x00100000 * numchips;
			break;			/* 1M * 4 = 4MB		*/

		case INTEL_ID_28F320J3A:
			info->flash_id += FLASH_28F320J3A;
			info->sector_count = 32;
			info->size = 0x00400000 * numchips;
			break;			/* 4M * 4 = 4MB		*/

		case INTEL_ID_28F640J3A:
			info->flash_id += FLASH_28F640J3A;
			info->sector_count = 64;
			info->size = 0x00800000 * numchips;
			break;			/* 8M * 4 = 4MB		*/

		case INTEL_ID_28F128J3A:
			info->flash_id += FLASH_28F128J3A;
			info->sector_count = 128;
			info->size = 0x00800000 * numchips;
			break;			/* 16M * 4 = 4MB	*/

#if 0	/* enable when device IDs are available */
		case AMD_ID_LV320T:
			info->flash_id += FLASH_AM320T;
			info->sector_count = 67;
			info->size = 0x00200000 * numchips;
			break;			/* 2M * 4 = 8MB		*/

		case AMD_ID_LV320B:
			info->flash_id += FLASH_AM320B;
			info->sector_count = 67;
			info->size = 0x00200000 * numchips;
			break;			/* 2M * 4 = 8MB		*/
#endif
		default:
			printf("%s: unknown id %x\n", __FUNCTION__,id);
			info->flash_id = FLASH_UNKNOWN;
			return (0);		/* => no or unknown flash */

	}

#if 0
	/* set up sector start address table */
	if (info->flash_id & FLASH_BTYPE) {
		/* set sector offsets for bottom boot block type	*/
		info->start[0] = base + 0x00000000;
		info->start[1] = base + 0x00008000;
		info->start[2] = base + 0x0000C000;
		info->start[3] = base + 0x00010000;
		for (i = 4; i < info->sector_count; i++) {
			info->start[i] = base + (i * 0x00020000) - 0x00060000;
		}
	} else {
		/* set sector offsets for top boot block type		*/
		i = info->sector_count - 1;
		info->start[i--] = base + info->size - 0x00008000;
		info->start[i--] = base + info->size - 0x0000C000;
		info->start[i--] = base + info->size - 0x00010000;
		for (; i >= 0; i--) {
			info->start[i] = base + i * 0x00020000;
		}
	}
#else
	flash_get_offsets ((ulong)addr, info);
#endif

	/* check for protected sectors */
	for (i = 0; i < info->sector_count; i++)
	{
		/* read sector protection at sector address, (A7 .. A0) = 0x02 */
		/* D0 = 1 if protected */
		if((info->flash_id & FLASH_TYPEMASK) == FLASH_ROM) info->protect[i]=1;
		else if((info->flash_id & FLASH_TYPEMASK) == FLASH_RAM) info->protect[i]=0;
		else {
			addr = (volatile unsigned char *)(info->start[i]);
			info->protect[i] = addr[2] & 1;
		}
	}

	/*
	 * Prevent writes to uninitialized FLASH.
	 */
	if (info->flash_id != FLASH_UNKNOWN)
	{
		flash_cmd(info, (unsigned char *)info->start[0], 0, 0xF0);
	}

	return (info->size);
}


/*-----------------------------------------------------------------------
 */

int	flash_erase (flash_info_t *info, int s_first, int s_last)
{
	vu_char *addr = (vu_char*)(info->start[0]);
	int flag, prot, sect, l_sect;
	ulong start, now, last;

	if((info->flash_id & FLASH_TYPEMASK) == FLASH_ROM) return 0;
	if((info->flash_id & FLASH_TYPEMASK) == FLASH_RAM) return 0;

	if ((s_first < 0) || (s_first > s_last)) {
		if (info->flash_id == FLASH_UNKNOWN) {
			printf ("- missing\n");
		} else {
			printf ("- no sectors to erase\n");
		}
		return 1;
	}

	if ((info->flash_id&FLASH_VENDMASK) == FLASH_MAN_INTEL) {
	    return flash_erase_intel(info, s_first, s_last);
	}

	if ((info->flash_id == FLASH_UNKNOWN) ||
	    (info->flash_id > FLASH_AMD_COMP)) {
		printf ("Can't erase unknown flash type - aborted\n");
		return 1;
	}

	prot = 0;
	for (sect=s_first; sect<=s_last; ++sect) {
		if (info->protect[sect]) {
			prot++;
		}
	}

	if (prot) {
		printf ("- Warning: %d protected sectors will not be erased!\n",
			prot);
	} else {
		printf ("\n");
	}

	l_sect = -1;

	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts();

	flash_cmd(info, addr, 0x555, 0xaa);
	flash_cmd(info, addr, 0x2aa, 0x55);
	flash_cmd(info, addr, 0x555, 0x80);
	flash_cmd(info, addr, 0x555, 0xaa);
	flash_cmd(info, addr, 0x2aa, 0x55);

	/* Start erase on unprotected sectors */
	for (sect = s_first; sect<=s_last; sect++) {
		if (info->protect[sect] == 0) {	/* not protected */
			addr = (vu_char*)(info->start[sect]);
			flash_cmd(info, addr, 0, 0x30);
			l_sect = sect;
		}
	}

	/* re-enable interrupts if necessary */
	if (flag)
		enable_interrupts();

	/* wait at least 80us - let's wait 1 ms */
	udelay (1000);

	/*
	 * We wait for the last triggered sector
	 */
	if (l_sect < 0)
		goto DONE;

	start = get_timer (0);
	last  = start;
	addr = (vu_char*)(info->start[l_sect]);
	while (!flash_cmp(info, addr,0,0xFF))
	{
		if ((now = get_timer(start)) > CFG_FLASH_ERASE_TOUT) {
			printf ("Timeout\n");
			return 1;
		}
		/* show that we're waiting */
		if ((now - last) > 1000) {	/* every second */
			putc ('.');
			last = now;
		}
	}

DONE:
	flash_cmd(info, (unsigned char *)info->start[0], 0, 0xF0);

	printf (" done\n");
	return 0;
}

/*-----------------------------------------------------------------------
 * Copy memory to flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */

int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
	ulong cp, wp=addr, data;
	int i, l, rc;
	int pwidth=info->portwidth;

	if((info->flash_id & FLASH_TYPEMASK) == FLASH_ROM) return 0;
	if((info->flash_id & FLASH_TYPEMASK) == FLASH_RAM) {
	    memcpy((void *)addr, src, cnt);
	    return 0;
	}

	switch(pwidth) {
	    case 4: wp = (addr & ~3);	/* get lower word aligned address */
	    case 2: wp = (addr & ~1);	/* get lower short aligned address */
	}

	/*
	 * handle unaligned start bytes
	 */
	if ((l = addr - wp) != 0) {
		data = 0;
		for (i=0, cp=wp; i<l; ++i, ++cp) {
			data = (data << 8) | (*(uchar *)cp);
		}
		for (; i<pwidth && cnt>0; ++i) {
			data = (data << 8) | *src++;
			--cnt;
			++cp;
		}
		for (; cnt==0 && i<pwidth; ++i, ++cp) {
			data = (data << 8) | (*(uchar *)cp);
		}

		if ((rc = write_word(info, wp, data)) != 0) {
			return (rc);
		}
		wp += pwidth;
	}

	/*
	 * handle word aligned part
	 */
	while (cnt >= pwidth) {
		data = 0;
		for (i=0; i<pwidth; ++i) {
			data = (data << 8) | *src++;
		}
		if ((rc = write_word(info, wp, data)) != 0) {
			return (rc);
		}
		wp  += pwidth;
		cnt -= pwidth;
	}

	if (cnt == 0) {
		return (0);
	}

	/*
	 * handle unaligned tail bytes
	 */
	data = 0;
	for (i=0, cp=wp; i<pwidth && cnt>0; ++i, ++cp) {
		data = (data << 8) | *src++;
		--cnt;
	}
	for (; i<pwidth; ++i, ++cp) {
		data = (data << 8) | (*(uchar *)cp);
	}

	return (write_word(info, wp, data));
}

/*-----------------------------------------------------------------------
 * Write a word to Flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */
static int write_word (flash_info_t *info, ulong dest, ulong data)
{
	vu_char *addr = (vu_char*)(info->start[0]);
	ulong start;
	int flag;

    	/* 32 bit chipwidth unsupported */
	unsigned long bit7=(info->chipwidth==1)?0x80808080:0x00800080;

	/* Check if Flash is (sufficiently) erased */
	if ((rw(info->portwidth, dest) & data) != data) {
		printf("%s: not erased enough %08lx %08lx\n", __FUNCTION__,
			rw(info->portwidth,dest),data);
		return (2);
	}

	if((info->flash_id & FLASH_VENDMASK) == FLASH_MAN_INTEL) {
	    return write_word_intel(info, dest, data);
	}

	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts();

	flash_cmd(info, addr, 0x555, 0xaa);
	flash_cmd(info, addr, 0x2AA, 0x55);
	flash_cmd(info, addr, 0x555, 0xa0);

	ww(info->portwidth, dest, data);

	/* re-enable interrupts if necessary */
	if (flag)
		enable_interrupts();

	/* data polling for D7 */
	start = get_timer (0);
	while ((rw(info->portwidth, dest) & bit7) != (data & bit7))
	{
		if (get_timer(start) > CFG_FLASH_WRITE_TOUT) {
		        printf("%s: timeout %08lx %08lx\n", __FUNCTION__,
				rw(info->portwidth, dest), data);
			return (1);
		}
	}
	return (0);
}

/*-----------------------------------------------------------------------
 */
